from IPython.display import display, Math
from fractions import Fraction

from .magic_dictionary import magic_dictionary


def convert_value(value):
    if type(value) is float:
        # For float value, we don't handle float:
        raise TypeError('Cannot set value as float, use fractions.Fraction instead.')
    return Fraction(value)


class simplex_dictionary:

    """ Class representing a dictionary for the simplex algorithm. The class contains
    multiple members representing the elements of a dictionary.

    Members:
      - B The list of basic variables (immutable).
      - N The list of non-basic variables (immutable).
      - b The current value of the basic variables.
      - a The current coefficients of the non-basic variables in the expression
          of the basic variables.
      - c The current coefficients of the non-basic variables in the objective.
      - z The current value of the objective.

    The magic member ".variables" can also be used to access the list of all
    variables.

    The various arrays of coefficients or values (b, a, c) are indexed by their respective
    variables, which must be in B or N.

    All values are converted to fractions.Fraction object in order to maintain consistency
    in the dictionary.

    Examples:
    # Create a simplex dictionary with two basic and two non-basic variables:
    >>> sdict = simplex_dictionary(B=['x_1', 'x_2'], N=['x_3', 'x_4'])

    # Set the value of x_1:
    >>> sdict.b['x_1'] = 4

    # Set the coefficient of x_3 in x_1:
    >>> sdict.a['x_1']['x_3'] = 12
    """

    def __init__(self, B, N):
        """ Create a new simplex dictionary with the given basic and non-basic variables.

        Parameters:
          - B  The list of basic variables.
          - N  The list of non-basic variables.
        """
        self.__B = tuple(B)
        self.__N = tuple(N)
        self.b = {}
        self.a = {}
        self.c = {}
        self.z = 0

    def _check_basic(self, key):
        """ Check if the given key is a basic variable, returning None if it is,
        and a custom exception string if not. Suitable for magic_dictionary use. """
        if key not in self.B:
            return '{} is not a basic variable.'.format(key)
        return None

    def _check_non_basic(self, key):
        """ Check if the given key is a non-basic variable, returning None if it is,
        and a custom exception string if not. Suitable for magic_dictionary use. """
        if key not in self.N:
            return '{} is not a non-basic variable.'.format(key)
        return None

    @property
    def variables(self): return sorted(self.B + self.N)

    @property
    def B(self): return self.__B

    @property
    def N(self): return self.__N

    def __setattr__(self, key, value):

        # Base and non-basic:
        if key == 'b':
            value = magic_dictionary(
                value,
                key_predicate=self._check_basic,
                value_converter=convert_value)
        elif key == 'c':
            value = magic_dictionary(
                value,
                key_predicate=self._check_non_basic,
                value_converter=convert_value)
        elif key == 'a':
            value = magic_dictionary(
                value,
                key_predicate=self._check_basic,
                value_converter=lambda value: magic_dictionary(
                    value, key_predicate=self._check_non_basic,
                    value_converter=convert_value),
                default_factory=lambda: magic_dictionary(
                    key_predicate=self._check_non_basic,
                    value_converter=convert_value))
        elif key == 'z':
            value = convert_value(value)

        return super().__setattr__(key, value)

    def name_latex(self, name):
        """ Convert the given variable name. """
        name = str(name)

        s = name.split('_')

        # We only handle special case:
        if len(s) == 1 or len(s) > 2:
            return name

        return s[0] + '_{' + s[1] + '}'

    def value_latex(self, value):
        if value.denominator == 1:
            return str(value)
        return r'{}\frac{{{}}}{{{}}}'.format(
            '-' if value.numerator < 0 else '',
            abs(value.numerator), value.denominator)

    def display(self, name=None):
        """ Display this simplex dictionary on the standard Jupyter output.

        Parameters:
          - prefix  Name of the dictionary.
        """
        d = (r'\begin{{array}}{{r||{}}}'.format(
             'r|' * (1 + len(self.B)))
             + r' & b & ' + ' & '.join(self.name_latex(v) for v in self.N) + r'\\\hline '
             + r'\\'.join(
                    '{} & {} &'.format(
                        self.name_latex(b),
                        self.value_latex(self.b[b])
                    )
                    + ' & '.join(self.value_latex(-self.a[b][n]) for n in self.N)
                    for i, b in enumerate(self.B))
             + r'\\\hline\hline '
             + r'&'.join(['z', self.value_latex(self.z)] + [self.value_latex(self.c[n])
                                                            for n in self.N])
             + r'\\\hline\end{array}')

        if name is not None:
            d = r'{} = \left.{}\right.'.format(name, d)

        display(Math(d))
